use rand::{rngs::OsRng, RngCore};
use scrypt::{scrypt, Params};
use sea_orm::{entity::prelude::*, ActiveValue::Set, QuerySelect};

#[derive(Clone, Debug, PartialEq, DeriveEntityModel, Eq)]
#[sea_orm(table_name = "master_pass")]
pub struct Model {
    #[sea_orm(primary_key, auto_increment = false)]
    pub user_id: u64,
    #[sea_orm(column_type = "Binary(BlobSize::Blob(Some(64)))")]
    pub salt: Vec<u8>,
    #[sea_orm(column_type = "Binary(BlobSize::Blob(Some(64)))")]
    pub password_hash: Vec<u8>,
}

#[derive(Copy, Clone, Debug, EnumIter, DeriveRelation)]
pub enum Relation {}

impl ActiveModelBehavior for ActiveModel {}

/// Hashes the password with Scrypt with the given salt
#[inline]
fn hash_password(password: &[u8], salt: &[u8]) -> crate::Result<[u8; 64]> {
    let params = Params::new(14, Params::RECOMMENDED_R, Params::RECOMMENDED_P, 64)?;
    let mut password_hash = [0; 64];
    scrypt(password, salt, &params, &mut password_hash)?;
    Ok(password_hash)
}

impl Model {
    /// Checks that the given password hash matches the one of the model
    pub fn verify(&self, password: &str) -> crate::Result<bool> {
        let hashed = hash_password(password.as_bytes(), &self.salt)?;
        Ok(hashed == self.password_hash.as_slice())
    }
}

impl ActiveModel {
    /// Hashes the password and creates an ActiveModel with all fields set to Set variant
    #[inline]
    pub fn from_unencrypted(user_id: u64, password: &str) -> crate::Result<Self> {
        let mut salt = vec![0; 64];
        OsRng.fill_bytes(&mut salt);
        let password_hash = Set(hash_password(password.as_bytes(), &salt)?.to_vec());
        Ok(Self {
            user_id: Set(user_id),
            salt: Set(salt),
            password_hash,
        })
    }
}

impl Entity {
    /// Gets the master password from the database
    #[inline]
    pub async fn get(user_id: u64, db: &DatabaseConnection) -> crate::Result<Option<Model>> {
        Self::find_by_id(user_id).one(db).await.map_err(Into::into)
    }

    /// Checks if the master password for the user exists
    #[inline]
    pub async fn exists(user_id: u64, db: &DatabaseConnection) -> crate::Result<bool> {
        let id = Self::find_by_id(user_id)
            .select_only()
            .column(Column::UserId)
            .into_tuple::<u64>()
            .one(db)
            .await?;
        Ok(id.is_some())
    }

    /// Removes a master password of the user from the database
    pub async fn remove(user_id: u64, db: &DatabaseConnection) -> crate::Result<()> {
        Self::delete_by_id(user_id).exec(db).await?;
        Ok(())
    }
}
